(* module Hilevel *)

open Types

let small = function
  | _,TermVar _ -> true
  | _ -> false

let term_of_result = function
  | fi,Symbol ident -> fi,TermVar ident
  | fi,ResLambda(ident,body) -> fi,TermLambda(ident,body)
  | fi,Closure(ident,body,env) ->
      (* This is a hack to signal the printing routines they're
         printing a Closure. term_of_result isn't invoked in any
         other circumstances under VB. *)
      fi,TermLambda("",body)
  | fi,ResPrimfun(prim) -> fi,TermPrimfun(prim)
  | fi,ResConstructor(cons) -> fi,TermConstructor(cons)
  | fi,ResDestructor(dest) -> fi,TermDestructor(dest)
  | fi,ResLiteral(lit) -> fi,TermLiteral(lit)

let result_of_term = function
  | fi,TermVar(ident) -> fi,Symbol ident
  | fi,TermLambda(ident,body) ->
      (* this clause will never be invoked under VB *)
      fi,ResLambda(ident,body)
  | fi,TermPrimfun(prim) -> fi,ResPrimfun(prim)
  | fi,TermConstructor(cons) -> fi,ResConstructor(cons)
  | fi,TermDestructor(dest) -> fi,ResDestructor(dest)
  | fi,TermLiteral(lit) -> fi,ResLiteral(lit)
  | _ -> invalid_arg "result_of_term"


module Printing = struct

  include Lolevel.Printing

  let rec print_term outer = function
    | _,TermLambda(x,body) ->
        if x = ""
        then print "<closure>" (* see hack above, in term_of_result *)
        else let () = begin obox(); print "\\"; print x end in
        let rec aux = function
          | _,TermLambda(y,t) -> begin space(); print y; aux t end
          | _ as t -> t in
        let body' = aux body in
        begin
          print ".";
          if (small body') && not outer then break() else space();
          print_term outer body';
          cbox()
        end
    | _,TermLet(x,arg,body) -> begin
        obox0();
        print "let ";
        print x;
        print " = ";
        print_term false arg;
        space();
        print "in";
        space();
        print_term false body;
        cbox()
      end
    | _,TermLetRec(xts,body) -> failwith "letrec not implemented"
    | _,TermIf(test,yes,no) -> begin
        obox0();
        print "if ";
        print_term false test;
        space();
        print "then ";
        print_term false yes;
        space();
        print "else ";
        print_term false no;
        cbox()
      end
    | t -> print_appterm outer t

  and print_appterm outer = function
    | _,TermApp(head,arg) -> begin
        obox0();
        print_appterm false head;
        space();
        print_atom false arg;
        cbox()
      end
    | t -> print_atom outer t

  and print_atom outer =
    (* let hole_pr = fun () -> print "_" in *)
    let void_pr = fun () -> () in
    function
    | fi,TermVar x -> print x
    | _,TermPrimfun(Funct1(pr,_)) -> pr ()
    | _,TermPrimfun(Funct2(pr,_,_)) -> pr void_pr ()
    | _,TermConstructor(Cons0(pr,proj)) -> pr (Some proj)
    | _,TermConstructor(Cons1(pr,_,_)) -> pr void_pr None
    | _,TermConstructor(Cons2(pr,_,_,_)) -> pr void_pr void_pr None
    | _,TermConstructor(Cons3(pr,_,_,_,_)) ->pr void_pr void_pr void_pr None
    | _,TermDestructor d ->
        (match d with
        | Fst -> print "fst"
        | Snd -> print "snd"
        | Trd -> print "trd"
        | IsZero -> print "zero?"
        | Pred -> print "pred"
      (*
        | IsNil -> print "nil?"
        | Head -> print "head"
        | Tail -> print "tail"
        | _ -> failwith "won't print private destructors"
      *)
        )
    | _,TermLiteral l ->
        (match l with
        | Bool true -> print "true"
        | Bool false -> print "false"
        | Num n -> print_int n
        | Str s -> begin
            obox0();
            print ("\"" ^ s ^ "\"");
            cbox()
          end)
    | term -> begin
        print "(";
        print_term outer term;
        print ")"
      end

  let print_binding = function
    | LocalBind _ -> ()
    | TopBind res -> begin print "= "; print_term true (term_of_result res) end
    | SymbolBind -> ()

  let print_result outer res =
    print_atom outer (term_of_result res)

end (* module Printing *)

open Printing

let apply_prim fi arg = function
  | Funct1(_,gob1) -> gob1 arg
  | Funct2(pr,gob1,gob2) ->
      (match gob1 arg with
      | None -> None
      | Some res ->
          let argpr () = print_result false res in
          Some(fi,ResPrimfun(Funct1(pr argpr,gob2 res))))

let apply_cons fi arg = function
  | Cons0 _ -> None
  | Cons1(pr,proj,gob1) ->
      (match gob1 arg with
      | None -> None
      | Some res ->
          let argpr () = print_result false res in
          Some(fi,ResConstructor(Cons0(pr argpr,proj res))))
  | Cons2(pr,proj,gob1,gob2) ->
      (match gob1 arg with
      | None -> None
      | Some res ->
          let argpr () = print_result false res in
          Some(fi,ResConstructor(Cons1(pr argpr,proj res,gob2 res))))
  | Cons3(pr,proj,gob1,gob2,gob3) ->
      (match gob1 arg with
      | None -> None
      | Some res ->
          let argpr () = print_result false res in
          Some(fi,ResConstructor(Cons2(pr argpr,proj res,gob2 res,gob3 res))))

